import functools
import multiprocessing

from django.conf import settings

from multi_tenant.migration_executors.base import (
    MigrationExecutor,
    run_migrations
)


class ParallelExecutor(MigrationExecutor):
    codename = 'parallel'

    def run_tenant_migrations(self, tenants):
        if tenants:
            processes = getattr(
                settings, 'TENANT_PARALLEL_MIGRATION_MAX_PROCESSES', 2
            )
            chunks = getattr(settings, 'TENANT_PARALLEL_MIGRATION_CHUNKS', 2)

            from django.db import connection

            connection.close()
            connection.connection = None

            run_migrations_p = functools.partial(
                run_migrations,
                self.args,
                self.options,
                self.codename,
                allow_atomic=False
            )
            p = multiprocessing.Pool(processes=processes)
            p.map(run_migrations_p, tenants, chunks)



from multi_tenant.management.commands.migrate_schemas import (
    Command as MigrateSchemasCommand
)


class Command(MigrateSchemasCommand):
    """
    Override django migrate command to enable all
    tenant features to other integrations.
    """



from django.core.management.base import CommandError
from django.core.management.commands.migrate import Command as MigrateCommand
from django.db.migrations.exceptions import MigrationSchemaMissing

from multi_tenant.constants import ALLOWED_BACKENDS
from multi_tenant.management.commands import SyncCommon
from multi_tenant.migration_executors import get_executor
from multi_tenant.utils import (
    get_engine_by_database,
    get_public_schema_name,
    get_tenant_model,
    schema_exists
)


def _check_backend(database):
    db_engine = get_engine_by_database(database)
    if db_engine not in ALLOWED_BACKENDS:
        raise CommandError(
            f'migrate has been disabled, for database {database}.'
            ' Use migrate_schemas instead. Please read the documentation'
            ' if you don\'t know why you shouldn\'t call migrate directly!'
        )


class Command(SyncCommon):
    help = """
    Updates database schema. Manages both apps
    with migrations and those without.
    """
    requires_system_checks = []

    def __init__(self, stdout=None, stderr=None, no_color=False):
        """
        Changes the option_list to use the options
        from the wrapped migrate command.
        """
        super(Command, self).__init__(stdout, stderr, no_color)

    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)
        command = MigrateCommand()
        command.add_arguments(parser)

    def handle(self, *args, **options):
        database = options.get('database', 'default')
        _check_backend(database=database)

        super(Command, self).handle(*args, **options)

        self.PUBLIC_SCHEMA_NAME = get_public_schema_name()

        executor = get_executor(codename=self.executor)(
            self.args, self.options
        )

        if self.sync_public:
            self.schema_name = self.schema_name or self.PUBLIC_SCHEMA_NAME
            executor.run_migrations(tenants=[self.schema_name])

        if not self.sync_tenant:
            return

        TenantModel = get_tenant_model()

        if self.schema_name and self.schema_name != self.PUBLIC_SCHEMA_NAME:
            if not schema_exists(self.schema_name):
                raise MigrationSchemaMissing(
                    f'Schema {self.schema_name} does not exist.'
                )

            tenants_qs = TenantModel.objects.filter(
                schema_name=self.schema_name
            )
        else:
            tenants_qs = TenantModel.objects.exclude(
                schema_name=self.PUBLIC_SCHEMA_NAME
            )

        tenants_qs = tenants_qs.select_related('tenant')
        executor.run_migrations(tenants=tenants_qs)



from django.conf import settings
from django.db.models.base import ModelBase
from django.db.utils import load_backend

from multi_tenant.constants import ALLOWED_BACKENDS
from multi_tenant.utils import get_engine_by_database, get_tenant_model


class TenantSyncRouter(object):
    """
    A router to control which applications will be synced,
    depending if we are syncing the shared apps or the tenant apps.
    """

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        # the imports below need to be done here else django <1.5 goes crazy
        # https://code.djangoproject.com/ticket/20704
        from django.db import connection

        from multi_tenant.postgresql_backend.base import (
            DatabaseWrapper as TenantDbWrapper
        )
        from multi_tenant.utils import app_labels, get_public_schema_name

        db_engine = get_engine_by_database(db)
        if not (
            db_engine in ALLOWED_BACKENDS
            or issubclass(
                getattr(load_backend(db_engine), 'DatabaseWrapper'),
                TenantDbWrapper
            )
        ):
            return None

        if isinstance(app_label, ModelBase):
            # In django <1.7 the `app_label` parameter is actually `model`
            app_label = app_label._meta.app_label

        # TODO apps labels cache y buscar mas rapido no asi
        if connection.schema_name == get_public_schema_name():
            if app_label not in app_labels(settings.SHARED_APPS):
                return False
        else:
            if isinstance(connection.tenant, get_tenant_model()):
                if connection.tenant.is_leading():
                    if app_label not in app_labels(settings.TENANT_APPS):
                        return False
                elif app_label not in app_labels(settings.TENANT_DETAIL_APPS):
                    return False

        return None
